"""
Some utility functions.
"""
from __future__ import annotations

import dataclasses
import json
import logging
import os
import sys
import uuid
from collections import deque
from datetime import datetime, timedelta
from enum import Enum
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt
from scipy.spatial import distance_matrix

from overcooked_simulator import ROOT_DIR

if TYPE_CHECKING:
    from overcooked_simulator.counters import Counter
from overcooked_simulator.player import Player


def create_init_env_time():
    """Init time of the environment time, because all environments should have the same internal time."""
    return datetime(
        year=2000, month=1, day=1, hour=0, minute=0, second=0, microsecond=0
    )


def get_closest(point: npt.NDArray[float], counters: list[Counter]):
    """Determines the closest counter for a given 2d-coordinate point in the env.

    Args:
        point: The point in the env for which to find the closest counter
        counters: List of objects with a `pos` attribute to compare to.

    Returns: The closest counter for the given point.
    """

    return counters[
        np.argmin(distance_matrix([point], [counter.pos for counter in counters])[0])
    ]


def get_collided_players(
    player_idx, players: list[Player], player_radius: float
) -> list[Player]:
    player_positions = np.array([p.pos for p in players], dtype=float)
    distances = distance_matrix(player_positions, player_positions)[player_idx]
    player_radiuses = np.array([player_radius for p in players], dtype=float)
    collisions = distances <= player_radiuses + player_radius
    collisions[player_idx] = False

    return [players[idx] for idx, val in enumerate(collisions) if val]


def get_touching_counters(target: Counter, counters: list[Counter]) -> list[Counter]:
    return list(
        filter(
            lambda counter: np.linalg.norm(counter.pos - target.pos) == 1.0, counters
        )
    )


def find_item_on_counters(item_uuid: str, counters: list[Counter]) -> Counter | None:
    for counter in counters:
        if counter.occupied_by:
            if isinstance(counter.occupied_by, deque):
                for item in counter.occupied_by:
                    if item.uuid == item_uuid:
                        return counter
            else:
                if item_uuid == counter.occupied_by.uuid:
                    return counter


def custom_asdict_factory(data):
    """Convert enums to their value."""

    def convert_value(obj):
        if isinstance(obj, Enum):
            return obj.value
        return obj

    return dict((k, convert_value(v)) for k, v in data)


def setup_logging(enable_websocket_logging=False):
    path_logs = ROOT_DIR.parent / "logs"
    os.makedirs(path_logs, exist_ok=True)
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s %(levelname)-8s %(name)-50s %(message)s",
        handlers=[
            logging.FileHandler(
                path_logs / f"{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}_debug.log",
                encoding="utf-8",
            ),
            logging.StreamHandler(sys.stdout),
        ],
    )
    logging.getLogger("matplotlib").setLevel(logging.WARNING)
    if not enable_websocket_logging:
        logging.getLogger("asyncio").setLevel(logging.ERROR)
        logging.getLogger("asyncio.coroutines").setLevel(logging.ERROR)
        logging.getLogger("websockets.server").setLevel(logging.ERROR)
        logging.getLogger("websockets.protocol").setLevel(logging.ERROR)
        logging.getLogger("websockets.client").setLevel(logging.ERROR)


def url_and_port_arguments(parser, server_name="game server", default_port=8000):
    parser.add_argument(
        "-url",
        "--url",
        "--host",
        type=str,
        default="localhost",
        help=f"Overcooked {server_name} host url.",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=default_port,
        help=f"Port number for the {server_name}",
    )


def disable_websocket_logging_arguments(parser):
    parser.add_argument(
        "--enable-websocket-logging" "", action="store_true", default=True
    )


def add_list_of_manager_ids_arguments(parser):
    parser.add_argument(
        "-m",
        "--manager_ids",
        nargs="+",
        type=str,
        default=[uuid.uuid4().hex],
        help="List of manager IDs that can create environments.",
    )


class NumpyAndDataclassEncoder(json.JSONEncoder):
    """Special json encoder for numpy types"""

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, timedelta):
            return obj.total_seconds()
        elif isinstance(obj, datetime):
            return obj.isoformat()
        elif dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj, dict_factory=custom_asdict_factory)
        # elif callable(obj):
        #     return getattr(obj, "__name__", "Unknown")

        return json.JSONEncoder.default(self, obj)
